#pragma once

// C++ system level
#include <map>
#include <string>
#include <cstring>  // memcpy
#include <functional>

// ESP32 specific
// #include "esp_log.h"

// project specific
#include "SpiderLib/NumberTypes.hpp"


// qthing stuff
// #include ""


namespace SpiderLib {

  // Forward declaration
  template <class>
  class CallbackMap;


  template <typename... Args>
  class CallbackMap<void(Args...)> {

    public:
      typedef std::function<void(Args...)> CallbackF;

      CallbackMap(u32 maxPrefixLength = 16) : maxPrefixLength(maxPrefixLength) {};

      // Doesn't work with zero-length prefixes (callback won't be called but no error occurs)
      void operator()(const u8* const buffer, u32 length, Args&&... args) const {

        u32 _len = std::min(maxPrefixLength, length);

        char _prefix[_len + 1];
        _prefix[_len] = 0x00;
        std::memcpy(_prefix, buffer, _len);
        std::string prefix(_prefix);

        u32 bestMatchLEN = 0;
        CallbackF callback = NULL;
        for (auto it = callbackMap.begin(); it != callbackMap.end(); ++it) {

          // input buffer is >= current iterators magic string length
          if (_len >= it->first.length() && prefix.find(it->first) == 0) {

            u32 matchLEN = it->first.length();
            if (matchLEN > bestMatchLEN) {
              bestMatchLEN = it->first.length();
              callback = it->second;
            }
          }
        }

        if (callback != NULL) callback(std::forward<Args>(args)...);

      }

      void add(const std::string& magicString, CallbackF callback) const {

        if (magicString.length() > maxPrefixLength) {
          // TODO: raise exception? log error?
          return;
        }

        if (callbackMap.count(magicString) > 0) {

          CallbackF oldCallback = callbackMap.at(magicString);

          callbackMap[magicString] = [oldCallback, callback](Args&&... args) {
            oldCallback(std::forward<Args>(args)...);
            callback(std::forward<Args>(args)...);
          };
        } else {
          callbackMap[magicString] = callback;
        }
      }

    private:
      u32 maxPrefixLength;
      mutable std::map<std::string, CallbackF> callbackMap;
  };

}
